"""Utility functions for CLI auto-completion functionality."""

# ruff: noqa: ARG001
import os
from contextlib import suppress
from pathlib import Path
from typing import Any

import requests.exceptions
from click import Context
from click.parser import Argument, split_arg_string
from click.shell_completion import CompletionItem
from cmem.cmempy.dp.admin import get_prometheus_data
from cmem.cmempy.dp.authorization.conditions import (
    fetch_all_acls,
    get_actions,
    get_groups,
    get_users,
)
from cmem.cmempy.health import get_complete_status_info
from cmem.cmempy.keycloak.client import list_open_id_clients
from cmem.cmempy.keycloak.group import list_groups
from cmem.cmempy.keycloak.user import get_user_by_username, list_users, user_groups
from cmem.cmempy.plugins.marshalling import get_marshalling_plugins
from cmem.cmempy.queries import QUERY_CATALOG
from cmem.cmempy.vocabularies import get_vocabularies
from cmem.cmempy.workflow.workflows import get_workflows_io
from cmem.cmempy.workspace import (
    get_task_plugin_description,
    get_task_plugins,
)
from cmem.cmempy.workspace.projects.datasets.dataset import get_dataset
from cmem.cmempy.workspace.projects.project import get_projects
from cmem.cmempy.workspace.projects.resources import get_all_resources
from cmem.cmempy.workspace.projects.variables import get_all_variables
from cmem.cmempy.workspace.python import list_packages
from cmem.cmempy.workspace.search import list_items
from natsort import natsorted, ns
from prometheus_client.parser import text_string_to_metric_families

from cmem_cmemc.constants import NS_ACL, NS_USER
from cmem_cmemc.context import CONTEXT
from cmem_cmemc.utils import (
    convert_iri_to_qname,
    get_graphs,
    get_published_packages,
    metric_get_labels,
    metrics_get_dict,
    struct_to_table,
)

SORT_BY_KEY = 0
SORT_BY_DESC = 1


def _finalize_completion(
    candidates: list,
    incomplete: str = "",
    sort_by: int = SORT_BY_KEY,
    nat_sort: bool = False,
    reverse: bool = False,
) -> list[CompletionItem]:
    """Sort and filter candidates list.

    candidates are sorted with natsort library by sort_by key.

    Args:
    ----
        candidates (list):  completion dictionary to filter
        incomplete (str):   incomplete string at the cursor
        sort_by (str):      SORT_BY_KEY or SORT_BY_DESC
        nat_sort (bool):    if true, uses the natsort package for sorting
        reverse (bool):     if true, sorts in reverse order

    Returns:
    -------
        filtered and sorted candidates list

    Raises:
    ------
        ValueError in case of wrong sort_by parameter

    """
    if sort_by not in (SORT_BY_KEY, SORT_BY_DESC):
        raise ValueError("sort_by should be 0 or 1.")
    incomplete = incomplete.lower()
    if len(candidates) == 0:
        return candidates
    # remove duplicates
    candidates = list(set(candidates))
    if isinstance(candidates[0], str):
        # list of strings filtering and sorting
        filtered_candidates = [
            element for element in candidates if element.lower().find(incomplete) != -1
        ]
        if nat_sort:
            return natsorted(seq=filtered_candidates, alg=ns.IGNORECASE, reverse=reverse)
        # this solves that case-insensitive sorting is not stable in ordering
        # of "equal" keys (https://stackoverflow.com/a/57923460)
        return sorted(filtered_candidates, key=lambda x: (str(x).casefold(), x), reverse=reverse)
    if isinstance(candidates[0], tuple):
        # list of tuples filtering and sorting
        filtered_candidates = [
            element
            for element in candidates
            if str(element[0]).lower().find(incomplete) != -1
            or str(element[1]).lower().find(incomplete) != -1
        ]
        if nat_sort:
            sorted_list = natsorted(
                seq=filtered_candidates,
                key=lambda k: k[sort_by],
                alg=ns.IGNORECASE,
                reverse=reverse,
            )
        else:
            sorted_list = sorted(
                filtered_candidates,
                key=lambda x: (str(x[sort_by]).casefold(), str(x[sort_by])),
                reverse=reverse,
            )
        return [
            CompletionItem(value=element[0].replace(":", r"\:"), help=element[1])
            for element in sorted_list
        ]
    raise ValueError(
        "Candidates should be a list of strings or a list of tuples." f" Got {candidates}"
    )


def _get_completion_args(incomplete: str) -> list[str]:
    """Get completion args

    This is a workaround to get partial tuple options in a completion function
    see https://github.com/pallets/click/issues/2597
    """
    args = split_arg_string(os.environ["COMP_WORDS"])
    if incomplete and len(args) > 0 and args[len(args) - 1] == incomplete:
        args.pop()
    return args


def _check_option_in_params(option: str, params: Any) -> bool:  # noqa: ANN401
    """Check if the given 'option' is present in the 'params' dictionary or any of its values."""
    if hasattr(params, "__iter__") and option in params:
        return True
    if option == params:
        return True
    return False


def add_metadata_parameter(list_: list | None = None) -> list:
    """Extend a list with metadata keys and key descriptions."""
    if list_ is None:
        list_ = []
    list_.insert(0, ("description", "Metadata: A description."))
    list_.insert(0, ("label", "Metadata: A name."))
    return list_


def acl_ids(ctx: Context, param: Argument, incomplete: str) -> list[CompletionItem]:
    """Prepare a list of access condition identifier"""
    CONTEXT.set_connection_from_params(ctx.find_root().params)
    options = []
    acls = fetch_all_acls()
    for access_condition in acls:
        iri = convert_iri_to_qname(access_condition.get("iri"), default_ns=NS_ACL)
        label = access_condition.get("name")
        if _check_option_in_params(iri, ctx.params.get(param.name)):  # type: ignore[attr-defined]
            continue
        options.append((iri, label))
    return _finalize_completion(candidates=options, incomplete=incomplete, sort_by=SORT_BY_DESC)


def acl_actions(ctx: Context, param: Argument, incomplete: str) -> list[CompletionItem]:
    """Prepare a list of access condition actions"""
    CONTEXT.set_connection_from_params(ctx.find_root().params)
    options = []
    results = get_actions().json()
    for _ in results:
        if _check_option_in_params(_["iri"], ctx.params.get(param.name)):  # type: ignore[attr-defined]
            continue
        options.append((_["iri"], _["name"]))
    return _finalize_completion(candidates=options, incomplete=incomplete, sort_by=SORT_BY_DESC)


def acl_users(ctx: Context, param: Argument, incomplete: str) -> list[CompletionItem]:
    """Prepare a list of access condition users"""
    CONTEXT.set_connection_from_params(ctx.find_root().params)
    options = []
    try:
        for _ in list_users():
            if _check_option_in_params(_["username"], ctx.params.get(param.name)):  # type: ignore[attr-defined]
                continue
            options.append(_["username"])
    except requests.exceptions.HTTPError:
        pass
    results = get_users().json()
    for _ in results:
        username = _.replace(NS_USER, "")
        if _check_option_in_params(username, ctx.params.get(param.name)) or username in options:  # type: ignore[attr-defined]
            continue
        options.append(username)

    return _finalize_completion(candidates=options, incomplete=incomplete, sort_by=SORT_BY_DESC)


def acl_groups(ctx: Context, param: Argument, incomplete: str) -> list[CompletionItem]:
    """Prepare a list of access condition groups"""
    CONTEXT.set_connection_from_params(ctx.find_root().params)
    options = []
    try:
        for _ in list_groups():
            if _check_option_in_params(_["name"], ctx.params.get(param.name)):  # type: ignore[attr-defined]
                continue
            options.append(_["name"])
    except requests.exceptions.HTTPError:
        pass
    results = get_groups().json()
    for _ in results:
        if _check_option_in_params(_, ctx.params.get(param.name)) or _ in options:  # type: ignore[attr-defined]
            continue
        options.append(_)
    return _finalize_completion(candidates=options, incomplete=incomplete, sort_by=SORT_BY_DESC)


def add_read_only_and_uri_property_parameters(list_: list | None = None) -> list:
    """Extend a list with readonly/uriProperty keys and key descriptions."""
    if list_ is None:
        list_ = []
    list_.append(
        (
            "readOnly",
            "Read-only: If enabled, all write operations using this dataset object "
            "will fail, e.g. when used as output in workflows or transform/linking "
            "executions. This will NOT protect the underlying resource in general, "
            "e.g. files, databases or knowledge graphs could still be changed "
            "externally.",
        )
    )
    list_.append(
        (
            "uriProperty",
            "URI attribute: When reading data from the dataset, the specified "
            "attribute will be used to get the URIs of the entities. "
            "When writing to a dataset, the specified attribute will be automatically "
            "added to the schema as well as the generated entity URIs will be added as "
            "values for each entity. If the entered value is not a valid URI, "
            "it will be converted to a valid URI.",
        )
    )
    return list_


def dataset_parameter(ctx: Context, param: Argument, incomplete: str) -> list[CompletionItem]:
    """Prepare a list of dataset parameters for a dataset type."""
    CONTEXT.set_connection_from_params(ctx.find_root().params)
    args = _get_completion_args(incomplete)
    incomplete = incomplete.lower()
    # look if cursor is in value position of the -p option and
    # return nothing in case it is (values are not completed atm)
    if args[len(args) - 2] in ("-p", "--parameter"):
        return []
    # try to determine the dataset type
    dataset_type = ctx.params.get("dataset_type")
    if dataset_type is None:
        try:
            dataset_id = ctx.args[0]
            project = get_dataset(
                project_name=dataset_id.split(":")[0], dataset_name=dataset_id.split(":")[1]
            )
            dataset_type = project["data"]["type"]
        except IndexError:
            pass

    # without type, we know nothing
    if dataset_type is None:
        return []
    plugin = get_task_plugin_description(dataset_type)
    properties = plugin["properties"]
    options = []
    for key in properties:
        title = properties[key]["title"]
        description = properties[key]["description"]
        option = f"{title}: {description}"
        options.append((key, option))

    options = add_read_only_and_uri_property_parameters(options)
    # sorting: metadata on top, then parameter per key
    options = sorted(options, key=lambda k: k[0].lower())
    options = add_metadata_parameter(options)
    # restrict to search
    options = [
        key
        for key in options
        if (
            key[0].lower().find(incomplete.lower()) != -1
            or key[1].lower().find(incomplete.lower()) != -1
        )
    ]

    return [CompletionItem(value=option[0], help=option[1]) for option in options]


def dataset_types(ctx: Context, param: Argument, incomplete: str) -> list[CompletionItem]:
    """Prepare a list of dataset types."""
    CONTEXT.set_connection_from_params(ctx.find_root().params)
    incomplete = incomplete.lower()
    options = []
    plugins = get_task_plugins()
    for plugin_id in plugins:
        plugin = plugins[plugin_id]
        title = plugin["title"]
        description = plugin["description"].partition("\n")[0]
        option = f"{title}: {description}"
        if plugin["taskType"] == "Dataset" and (
            plugin_id.lower().find(incomplete.lower()) != -1
            or option.lower().find(incomplete.lower()) != -1
        ):
            options.append((plugin_id, option))
    return _finalize_completion(candidates=options, incomplete=incomplete, sort_by=SORT_BY_DESC)


def dataset_ids(ctx: Context, param: Argument, incomplete: str) -> list[CompletionItem]:
    """Prepare a list of projectid:datasetid dataset identifier."""
    CONTEXT.set_connection_from_params(ctx.find_root().params)
    results = list_items(item_type="dataset")
    datasets = results["results"]
    options = [(f"{_['projectId']}:{_['id']}", _["label"]) for _ in datasets]
    return _finalize_completion(candidates=options, incomplete=incomplete)


def dataset_list_filter(ctx: Context, param: Argument, incomplete: str) -> list[CompletionItem]:
    """Prepare a list of filter names and values for dataset list filter."""
    filter_names = [
        ("project", "Filter by project ID."),
        ("regex", "Filter by regular expression on the dataset label."),
        ("tag", "Filter by tag label."),
        ("type", "Filter by dataset type."),
    ]
    filter_regex = [
        (r"^Final\:", "Example: Dataset label starts with 'Final:'."),
        (
            r"[12][0-9][0-9][0-9]-[0-9][0-9]-[0-9][0-9]",
            "Example: Dataset label contains a data-like string.",
        ),
    ]
    options = []
    args = _get_completion_args(incomplete)
    if args[len(args) - 1] == "--filter":
        options = _finalize_completion(candidates=filter_names, incomplete=incomplete)
    if args[len(args) - 1] == "type":
        options = dataset_types(ctx, param, incomplete)
    if args[len(args) - 1] == "project":
        options = project_ids(ctx, param, incomplete)
    if args[len(args) - 1] == "tag":
        options = tag_labels(ctx, param, incomplete, "dataset")
    if args[len(args) - 1] == "regex":
        options = _finalize_completion(candidates=filter_regex, incomplete=incomplete)
    return options


def resource_ids(ctx: Context, param: Argument, incomplete: str) -> list[CompletionItem]:
    """Prepare a list of projectid:resourceid resource identifier."""
    CONTEXT.set_connection_from_params(ctx.find_root().params)
    options = [_["id"] for _ in get_all_resources()]
    return _finalize_completion(candidates=options, incomplete=incomplete, sort_by=SORT_BY_DESC)


def scheduler_ids(ctx: Context, param: Argument, incomplete: str) -> list[CompletionItem]:
    """Prepare a list of projectid:schedulerid scheduler identifier."""
    CONTEXT.set_connection_from_params(ctx.find_root().params)
    options = []
    schedulers = list_items(
        item_type="task",
        facets=[{"facetId": "taskType", "keywordIds": ["Scheduler"], "type": "keyword"}],
    )["results"]
    for _ in schedulers:
        if _check_option_in_params(_["projectId"] + ":" + _["id"], ctx.params.get(param.name)):  # type: ignore[attr-defined]
            continue
        options.append((_["projectId"] + ":" + _["id"], _["label"]))
    return _finalize_completion(candidates=options, incomplete=incomplete, sort_by=SORT_BY_DESC)


def metric_ids(ctx: Context, param: Argument, incomplete: str) -> list[CompletionItem]:
    """Prepare a list of metric identifier."""
    CONTEXT.set_connection_from_params(ctx.find_root().params)
    data = get_prometheus_data().text
    options = [
        (family.name, family.documentation) for family in text_string_to_metric_families(data)
    ]
    return _finalize_completion(candidates=options, incomplete=incomplete, sort_by=SORT_BY_KEY)


def metric_label_filter(ctx: Context, param: Argument, incomplete: str) -> list[CompletionItem]:
    """Prepare a list of label name or values."""
    CONTEXT.set_connection_from_params(ctx.find_root().params)
    args = _get_completion_args(incomplete)
    incomplete = incomplete.lower()
    options: list[str] = []
    metric_id = ctx.args[0]
    labels = metric_get_labels(metrics_get_dict()[metric_id])
    if args[len(args) - 1] in "--filter":
        # we are in the name position
        options = list(labels.keys())
    if args[len(args) - 2] in "--filter":
        label_name = args[len(args) - 1]
        # we are in the value position
        options = labels[label_name]
    return _finalize_completion(candidates=options, incomplete=incomplete, sort_by=SORT_BY_KEY)


def vocabularies(
    ctx: Context, param: Argument, incomplete: str, filter_: str = "all"
) -> list[CompletionItem]:
    """Prepare a list of vocabulary graphs for auto-completion."""
    CONTEXT.set_connection_from_params(ctx.find_root().params)
    vocabs = get_vocabularies(filter_=filter_)
    options = []
    for _ in vocabs:
        url = _["iri"]
        if _check_option_in_params(url, ctx.params.get(param.name)):  # type: ignore[attr-defined]
            continue
        url = _["iri"]
        try:
            label = _["label"]["title"]
        except (KeyError, TypeError):
            label = "Vocabulary in graph " + url
        options.append((url, label))
    return _finalize_completion(candidates=options, incomplete=incomplete, sort_by=SORT_BY_DESC)


def installed_vocabularies(ctx: Context, param: Argument, incomplete: str) -> list[CompletionItem]:
    """Prepare a list of installed vocabulary graphs."""
    return vocabularies(ctx, param, incomplete, filter_="installed")


def installable_vocabularies(
    ctx: Context, param: Argument, incomplete: str
) -> list[CompletionItem]:
    """Prepare a list of installable vocabulary graphs."""
    return vocabularies(ctx, param, incomplete, filter_="installable")


def file_list(
    incomplete: str = "", suffix: str = "", description: str = "", prefix: str = ""
) -> list[CompletionItem]:
    """Prepare a list of files with specific parameter."""
    directory = str(Path().cwd())
    options = [
        (file_name, description)
        for file_name in os.listdir(directory)
        if (
            Path(file_name).exists() and file_name.endswith(suffix) and file_name.startswith(prefix)
        )
    ]
    return _finalize_completion(candidates=options, incomplete=incomplete, sort_by=SORT_BY_KEY)


def workflow_io_ids(ctx: Context, param: Argument, incomplete: str) -> list[CompletionItem]:
    """Prepare a list of io workflows."""
    CONTEXT.set_connection_from_params(ctx.find_root().params)
    options = []
    for _ in get_workflows_io():
        workflow_id = _["projectId"] + ":" + _["id"]
        label = _["label"]
        options.append((workflow_id, label))
    return _finalize_completion(candidates=options, incomplete=incomplete, sort_by=SORT_BY_DESC)


def replay_files(ctx: Context, param: Argument, incomplete: str) -> list[CompletionItem]:
    """Prepare a list of JSON replay files."""
    return file_list(incomplete=incomplete, suffix=".json", description="JSON query replay file")


def installed_package_names(ctx: Context, param: Argument, incomplete: str) -> list[CompletionItem]:
    """Prepare a list of installed packages."""
    CONTEXT.set_connection_from_args(ctx.find_root().params)
    options = [(_["name"], _["version"]) for _ in list_packages()]
    return _finalize_completion(candidates=options, incomplete=incomplete, sort_by=SORT_BY_KEY)


def published_package_names(ctx: Context, param: Argument, incomplete: str) -> list[CompletionItem]:
    """List of plugin packages scraped from pypi.org."""
    options = [(_.name, f"{_.version}: {_.description}") for _ in get_published_packages()]
    return _finalize_completion(candidates=options, incomplete=incomplete, sort_by=SORT_BY_KEY)


def python_package_files(ctx: Context, param: Argument, incomplete: str) -> list[CompletionItem]:
    """Prepare a list of acceptable python package files."""
    return file_list(
        incomplete=incomplete,
        suffix=".tar.gz",
        description="Plugin Python Package file",
        prefix="cmem-plugin-",
    )


def installable_packages(ctx: Context, param: Argument, incomplete: str) -> list[CompletionItem]:
    """Installable packages from files and pypi.org."""
    return python_package_files(ctx, param, incomplete) + published_package_names(
        ctx, param, incomplete
    )


def workflow_io_input_files(ctx: Context, param: Argument, incomplete: str) -> list[CompletionItem]:
    """Prepare a list of acceptable workflow io input files."""
    return (
        file_list(incomplete=incomplete, suffix=".csv", description="CSV Dataset resource")
        + file_list(incomplete=incomplete, suffix=".xml", description="XML Dataset resource")
        + file_list(incomplete=incomplete, suffix=".json", description="JSON Dataset resource")
        + file_list(incomplete=incomplete, suffix=".xlsx", description="Excel Dataset resource")
        + file_list(incomplete=incomplete, suffix=".txt", description="Text Dataset resource")
        + file_list(incomplete=incomplete, suffix=".zip", description="Multi CSV Dataset resource")
    )


def workflow_io_input_mimetypes(
    ctx: Context, param: Argument, incomplete: str
) -> list[CompletionItem]:
    """Prepare a list of acceptable workflow io input mimetypes."""
    return (
        file_list(incomplete=incomplete, suffix=".csv", description="CSV Dataset resource")
        + file_list(incomplete=incomplete, suffix=".xml", description="XML Dataset resource")
        + file_list(incomplete=incomplete, suffix=".json", description="JSON Dataset resource")
        + file_list(incomplete=incomplete, suffix=".xlsx", description="Excel Dataset resource")
        + file_list(incomplete=incomplete, suffix=".txt", description="Text Dataset resource")
        + file_list(incomplete=incomplete, suffix=".zip", description="Multi CSV Dataset resource")
    )


def workflow_io_output_files(
    ctx: Context, param: Argument, incomplete: str
) -> list[CompletionItem]:
    """Prepare a list of acceptable workflow io output files."""
    return (
        file_list(incomplete=incomplete, suffix=".csv", description="CSV Dataset resource")
        + file_list(incomplete=incomplete, suffix=".xml", description="XML Dataset resource")
        + file_list(incomplete=incomplete, suffix=".json", description="JSON Dataset resource")
        + file_list(incomplete=incomplete, suffix=".xlsx", description="Excel Dataset resource")
        + file_list(incomplete=incomplete, suffix=".ttl", description="RDF file Dataset resource")
        + file_list(incomplete=incomplete, suffix=".nt", description="RDF file Dataset resource")
    )


def dataset_files(ctx: Context, param: Argument, incomplete: str) -> list[CompletionItem]:
    """Prepare a list of SPARQL files."""
    return (
        file_list(incomplete=incomplete, suffix=".csv", description="CSV Dataset resource")
        + file_list(incomplete=incomplete, suffix=".xlsx", description="Excel Dataset resource")
        + file_list(incomplete=incomplete, suffix=".xml", description="XML Dataset resource")
        + file_list(incomplete=incomplete, suffix=".json", description="JSON Dataset resource")
        + file_list(
            incomplete=incomplete, suffix=".jsonl", description="JSON Lines Dataset resource"
        )
        + file_list(incomplete=incomplete, suffix=".ttl", description="RDF file Dataset resource")
        + file_list(incomplete=incomplete, suffix=".zip", description="multiCsv Dataset resource")
        + file_list(incomplete=incomplete, suffix=".orc", description="Apache ORC Dataset resource")
        + file_list(incomplete=incomplete, suffix=".yaml", description="YAML Document")
        + file_list(incomplete=incomplete, suffix=".yml", description="YAML Document")
    )


def graph_backup_files(ctx: Context, param: Argument, incomplete: str) -> list[CompletionItem]:
    """Prepare a list of workspace files."""
    return file_list(
        incomplete=incomplete,
        suffix=".graphs.zip",
        description="eccenca Corporate Memory graph backup file",
    )


def project_files(ctx: Context, param: Argument, incomplete: str) -> list[CompletionItem]:
    """Prepare a list of workspace files."""
    return file_list(
        incomplete=incomplete,
        suffix=".project.zip",
        description="eccenca Corporate Memory project backup file",
    )


def ini_files(ctx: Context, param: Argument, incomplete: str) -> list[CompletionItem]:
    """Prepare a list of workspace files."""
    return file_list(incomplete=incomplete, suffix=".ini", description="INI file")


def workspace_files(ctx: Context, param: Argument, incomplete: str) -> list[CompletionItem]:
    """Prepare a list of workspace files."""
    return file_list(
        incomplete=incomplete,
        suffix=".workspace.zip",
        description="eccenca Corporate Memory workspace backup file",
    )


def sparql_files(ctx: Context, param: Argument, incomplete: str) -> list[CompletionItem]:
    """Prepare a list of SPARQL files."""
    return file_list(
        incomplete=incomplete, suffix=".sparql", description="SPARQL query file"
    ) + file_list(incomplete=incomplete, suffix=".rq", description="SPARQL query file")


def triple_files(ctx: Context, param: Argument, incomplete: str) -> list[CompletionItem]:
    """Prepare a list of triple files."""
    return file_list(
        incomplete=incomplete, suffix=".ttl", description="RDF Turtle file"
    ) + file_list(incomplete=incomplete, suffix=".nt", description="RDF NTriples file")


def placeholder(ctx: Context, param: Argument, incomplete: str) -> list[CompletionItem]:
    """Prepare a list of placeholder from the to-be executed queries."""
    # look if cursor is in value position of the -p option and
    # return nothing in case it is (values are not completed atm)
    args = _get_completion_args(incomplete)
    if args[len(args) - 2] in ("-p", "--parameter"):
        return []
    # setup configuration
    CONTEXT.set_connection_from_params(ctx.find_root().params)
    # extract placeholder from given queries in the command line
    options = []
    for _, arg in enumerate(args):
        query = QUERY_CATALOG.get_query(arg)
        if query is not None:
            options.extend(list(query.get_placeholder_keys()))
    # look for already given parameter in the arguments and remove them from
    # the available options
    for num, arg in enumerate(args):
        if num - 1 > 0 and args[num - 1] in ("-p", "--parameter"):
            options.remove(arg)
    return _finalize_completion(candidates=options, incomplete=incomplete)


def remote_queries(ctx: Context, param: Argument, incomplete: str) -> list[CompletionItem]:
    """Prepare a list of query URIs."""
    CONTEXT.set_connection_from_params(ctx.find_root().params)
    options = []
    for query in QUERY_CATALOG.get_queries().values():
        url = query.short_url
        label = query.label
        options.append((url, label))
    return _finalize_completion(candidates=options, incomplete=incomplete, sort_by=SORT_BY_DESC)


def remote_queries_and_sparql_files(
    ctx: Context, param: Argument, incomplete: str
) -> list[CompletionItem]:
    """Prepare a list of named queries, query files and directories."""
    remote = remote_queries(ctx, param, incomplete)
    files = sparql_files(ctx, param, incomplete)
    return remote + files


def workflow_ids(ctx: Context, param: Argument, incomplete: str) -> list[CompletionItem]:
    """Prepare a list of projectid:taskid workflow identifier."""
    CONTEXT.set_connection_from_params(ctx.find_root().params)
    workflows = list_items(item_type="workflow")["results"]
    options = []
    for _ in workflows:
        workflow = _["projectId"] + ":" + _["id"]
        label = _["label"]
        if _check_option_in_params(workflow, ctx.params.get(param.name)):  # type: ignore[attr-defined]
            continue
        options.append((workflow, label))
    return _finalize_completion(candidates=options, incomplete=incomplete, sort_by=SORT_BY_DESC)


def marshalling_plugins(ctx: Context, param: Argument, incomplete: str) -> list[CompletionItem]:
    """Prepare a list of supported workspace/project import/export plugins."""
    CONTEXT.set_connection_from_params(ctx.find_root().params)
    options = get_marshalling_plugins()
    if "description" in options[0]:
        final_options = [(_["id"], _["description"]) for _ in options]
    else:
        # in case, no descriptions are available, labels are fine as well
        final_options = [(_["id"], _["label"]) for _ in options]

    return _finalize_completion(
        candidates=final_options, incomplete=incomplete, sort_by=SORT_BY_DESC
    )


def project_ids(ctx: Context, param: Argument, incomplete: str) -> list[CompletionItem]:
    """Prepare a list of project IDs for auto-completion."""
    CONTEXT.set_connection_from_params(ctx.find_root().params)
    projects = get_projects()
    options = []
    for _ in projects:
        project_id = _["name"]
        label = _["metaData"]["label"]
        # do not add project if already in the command line
        if _check_option_in_params(project_id, ctx.params.get(param.name)):  # type: ignore[attr-defined]
            continue
        options.append((project_id, label))
    return _finalize_completion(candidates=options, incomplete=incomplete, sort_by=SORT_BY_DESC)


def graph_uris(
    ctx: Context, param: Argument, incomplete: str, writeable: bool = True, readonly: bool = True
) -> list[CompletionItem]:
    """Prepare a list of graphs for auto-completion."""
    CONTEXT.set_connection_from_params(ctx.find_root().params)
    graphs = get_graphs()
    options = []
    for _ in graphs:
        iri = _["iri"]
        label = _["label"]["title"]
        # do not add graph if already in the command line
        if _check_option_in_params(iri, ctx.params.get(param.name)):  # type: ignore[attr-defined]
            continue
        options.append((iri, label))
    return _finalize_completion(candidates=options, incomplete=incomplete, sort_by=SORT_BY_DESC)


def writable_graph_uris(ctx: Context, param: Argument, incomplete: str) -> list[CompletionItem]:
    """Prepare a list of writable graphs for auto-completion."""
    return graph_uris(ctx, param, incomplete, writeable=True, readonly=False)


def graph_uris_with_all_graph_uri(
    ctx: Context, param: Argument, incomplete: str
) -> list[CompletionItem]:
    """Please a list of all graphs for acl command auto-completion."""
    options = graph_uris(ctx, param, incomplete, writeable=True, readonly=True)
    options.append(CompletionItem(value=r"urn\:elds-backend-all-graphs", help="All Graphs"))
    return options


def connections(ctx: Context, param: Argument, incomplete: str) -> list[CompletionItem]:
    """Prepare a list of config connections for auto-completion."""
    # since ctx does not have an obj here, we re-create the object
    CONTEXT.set_connection_from_params(ctx.find_root().params)
    options = CONTEXT.config.sections()
    return _finalize_completion(candidates=options, incomplete=incomplete)


def graph_export_templates(ctx: Context, param: Argument, incomplete: str) -> list[CompletionItem]:
    """Prepare a list of example templates for the graph export command."""
    examples = [
        ("{{hash}}", "Example: 6568a[...]00b08.ttl"),
        ("{{iriname}}", "Example: https__ns_eccenca_com_data_config.ttl"),
        ("{{date}}-{{iriname}}", "Example: 2021-11-29-https__ns_eccenca_com_data_config.ttl"),
        (
            "{{date}}-{{connection}}-{{iriname}}",
            "Example: 2021-11-29-mycmem-https__ns_eccenca_com_data_config.ttl",
        ),
    ]
    return _finalize_completion(candidates=examples, incomplete=incomplete)


def project_export_templates(
    ctx: Context, param: Argument, incomplete: str
) -> list[CompletionItem]:
    """Prepare a list of example templates for the project export command."""
    examples = [
        ("{{id}}", "Example: a plain file name"),
        ("{{date}}-{{connection}}-{{id}}.project", "Example: a more descriptive file name"),
        ("dumps/{{connection}}/{{id}}/{{date}}.project", "Example: a whole directory tree"),
    ]
    return _finalize_completion(candidates=examples, incomplete=incomplete)


def workspace_export_templates(
    ctx: Context, param: Argument, incomplete: str
) -> list[CompletionItem]:
    """Prepare a list of example templates for the workspace export command."""
    examples = [
        ("workspace", "Example: a plain file name"),
        ("{{date}}-{{connection}}.workspace", "Example: a more descriptive file name"),
        ("dumps/{{connection}}/{{date}}.workspace", "Example: a whole directory tree"),
    ]
    return _finalize_completion(candidates=examples, incomplete=incomplete)


def graph_list_filter(ctx: Context, param: Argument, incomplete: str) -> list[CompletionItem]:
    """Prepare a list of filter names and values for graph list filter."""
    filter_names = [
        (
            "access",
            "List only graphs which have a certain access condition " "(readonly or writeable).",
        ),
        ("imported-by", "List only graphs which are in the import tree of a " "specified graph."),
    ]
    filter_values_access = [
        ("readonly", "List only graphs which are NOT writable for the current user."),
        ("writeable", "List only graphs which ARE writeable for the current user."),
    ]
    args = _get_completion_args(incomplete)
    options = []
    if args[len(args) - 1] == "--filter":
        options = _finalize_completion(candidates=filter_names, incomplete=incomplete)
    if args[len(args) - 1] == "access":
        options = _finalize_completion(candidates=filter_values_access, incomplete=incomplete)
    if args[len(args) - 1] == "imported-by":
        options = graph_uris(ctx, param, incomplete)
    return options


def variable_ids(ctx: Context, param: Argument, incomplete: str) -> list[CompletionItem]:
    """Prepare a list of variables IDs for auto-completion."""
    CONTEXT.set_connection_from_params(ctx.find_root().params)
    variables = get_all_variables()
    options = []
    for _ in variables:
        variable_id = _["id"]
        label = _.get("description", "").partition("\n")[0]
        if label == "":
            label = f"Current value: {_['value']}"
        # do not add project if already in the command line
        if _check_option_in_params(variable_id, ctx.params.get(param.name)):  # type: ignore[attr-defined]
            continue
        options.append((variable_id, label))
    return _finalize_completion(candidates=options, incomplete=incomplete, sort_by=SORT_BY_KEY)


def variable_list_filter(ctx: Context, param: Argument, incomplete: str) -> list[CompletionItem]:
    """Prepare a list of filter names and values for variable list filter."""
    filter_names = [
        ("project", "Filter for variables from a specific project."),
        (
            "regex",
            "Filter for variables with a regular expression search over "
            "id, value and description.",
        ),
    ]
    filter_values_regex = [
        ("ending$", "Variables name ends with 'ending'."),
        ("^starting", "Variables name starts with 'starting'."),
    ]
    args = _get_completion_args(incomplete)
    if args[len(args) - 1] == "--filter":
        return [CompletionItem(value=f[0], help=f[1]) for f in filter_names]
    if args[len(args) - 1] == "regex":
        return _finalize_completion(candidates=filter_values_regex, incomplete=incomplete)
    if args[len(args) - 1] == "project":
        return project_ids(ctx, param, incomplete)
    return []


def resource_list_filter(ctx: Context, param: Argument, incomplete: str) -> list[CompletionItem]:
    """Prepare a list of filter names and values for resource list filter."""
    filter_names = [
        ("project", "Filter for file resources from a specific project."),
        (
            "regex",
            "Filter for file resources with a regular expression search on " "the name field.",
        ),
    ]
    filter_values_regex = [
        ("csv$", "File resources which name ends with .csv"),
        ("2021-10-[0-9][0-9]", "File resources which name has a date from 2021-10 in it"),
    ]
    args = _get_completion_args(incomplete)
    if args[len(args) - 1] == "--filter":
        return [CompletionItem(value=f[0], help=f[1]) for f in filter_names]
    if args[len(args) - 1] == "project":
        return _finalize_completion(
            candidates=project_ids(ctx, param, incomplete), incomplete=incomplete
        )
    if args[len(args) - 1] == "regex":
        return _finalize_completion(candidates=filter_values_regex, incomplete=incomplete)
    return []


def workflow_list_filter(ctx: Context, param: Argument, incomplete: str) -> list[CompletionItem]:
    """Prepare a list of filter names and values for workflow list filter."""
    filter_names = [
        ("project", "Filter by project ID."),
        ("io", "Filter by workflow io feature."),
        ("regex", "Filter by regular expression on the workflow label."),
        ("tag", "Filter by tag label."),
    ]
    filter_values_io = [
        ("any", "List all workflows suitable for the io command."),
        ("input-only", "List only workflows with a variable input dataset."),
        ("output-only", "List only workflows with a variable output dataset."),
        ("input-output", "List only workflows with a variable input and output dataset."),
    ]
    filter_regex = [
        (r"^Final\:", "Example: Workflow label starts with 'Final:'."),
        (
            r"[12][0-9][0-9][0-9]-[0-9][0-9]-[0-9][0-9]",
            "Example: Workflow label contains a data-like string.",
        ),
    ]
    options = []
    args = _get_completion_args(incomplete)
    if args[len(args) - 1] == "--filter":
        options = filter_names
    if args[len(args) - 1] == "io":
        options = filter_values_io
    if args[len(args) - 1] == "project":
        return project_ids(ctx, param, incomplete)
    if args[len(args) - 1] == "tag":
        return tag_labels(ctx, param, incomplete, "workflow")
    if args[len(args) - 1] == "regex":
        options = filter_regex

    return _finalize_completion(candidates=options, incomplete=incomplete)


def tag_labels(
    ctx: Context, param: Argument, incomplete: str, item_type: str
) -> list[CompletionItem]:
    """Prepare a list of tag labels for a item_type."""
    datasets = list_items(item_type=item_type)
    options = []
    counts: dict[str, int] = {}
    for _dataset in datasets["results"]:
        for _tag in _dataset["tags"]:
            if _tag["label"] in counts:
                counts[_tag["label"]] += 1
            else:
                counts[_tag["label"]] = 1
    for tag, count in counts.items():
        options.append((tag, f"{count} item(s): {tag}"))
    return _finalize_completion(
        candidates=options, incomplete=incomplete, sort_by=SORT_BY_DESC, reverse=True
    )


def status_keys(ctx: Context, param: Argument, incomplete: str) -> list[CompletionItem]:
    """Prepare a list of status keys for the admin status command."""
    os.environ["CMEMPY_IS_CHATTY"] = "false"
    status_info = struct_to_table(get_complete_status_info())
    options = [_[0] for _ in status_info]
    options.insert(0, "all")
    return _finalize_completion(candidates=options, incomplete=incomplete)


def user_ids(ctx: Context, param: Argument, incomplete: str) -> list[CompletionItem]:
    """Prepare a list of username for admin update/delete/password command."""
    CONTEXT.set_connection_from_params(ctx.find_root().params)
    options = [_["username"] for _ in list_users()]
    return _finalize_completion(candidates=options, incomplete=incomplete, sort_by=SORT_BY_DESC)


def user_group_ids(ctx: Context, param: Argument, incomplete: str) -> list[CompletionItem]:
    """Prepare a list of group name for admin user update --(un)assign-group parameter"""
    CONTEXT.set_connection_from_params(ctx.find_root().params)
    if not ctx.args:
        return []
    users = get_user_by_username(username=str(ctx.args[0]))
    if not users:
        return []

    if param.name == "unassign_group":  # type: ignore[attr-defined]
        groups = user_groups(user_id=users[0]["id"])
    else:
        user_group_names = [group["name"] for group in user_groups(user_id=users[0]["id"])]
        groups = [group for group in list_groups() if group["name"] not in user_group_names]
    options = [_["name"] for _ in groups]

    for arg in ctx.params["assign_group"]:
        with suppress(ValueError):
            options.remove(arg)
    for arg in ctx.params["unassign_group"]:
        with suppress(ValueError):
            options.remove(arg)

    return _finalize_completion(candidates=options, incomplete=incomplete, sort_by=SORT_BY_DESC)


def client_ids(ctx: Context, param: Argument, incomplete: str) -> list[CompletionItem]:
    """Prepare a list of client ids for admin secret and update command."""
    CONTEXT.set_connection_from_params(ctx.find_root().params)
    options = [_["clientId"] for _ in list_open_id_clients()]
    return _finalize_completion(candidates=options, incomplete=incomplete, sort_by=SORT_BY_DESC)


def transformation_task_ids(ctx: Context, param: Argument, incomplete: str) -> list[CompletionItem]:
    """Prepare a list of projectId:transformation task identifier."""
    CONTEXT.set_connection_from_params(ctx.find_root().params)
    results = list_items(item_type="transform")
    datasets = results["results"]
    options = [(f"{_['projectId']}:{_['id']}", _["label"]) for _ in datasets]
    return _finalize_completion(candidates=options, incomplete=incomplete, sort_by=SORT_BY_DESC)


def linking_task_ids(ctx: Context, param: Argument, incomplete: str) -> list[CompletionItem]:
    """Prepare a list of projectId:linking task identifier."""
    CONTEXT.set_connection_from_args(ctx.find_root().params)
    results = list_items(item_type="linking")
    datasets = results["results"]
    options = [(_["projectId"] + r"\:" + _["id"], _["label"]) for _ in datasets]
    return _finalize_completion(candidates=options, incomplete=incomplete, sort_by=SORT_BY_DESC)



"""workflow commands for cmem command line interface."""

import re
import sys
import time
from datetime import datetime, timezone
from pathlib import Path

import click
import timeago
from click import UsageError
from cmem.cmempy.workflow import get_resource_based_dataset_types, get_workflows
from cmem.cmempy.workflow.workflow import (
    execute_workflow_io,
    get_workflow_editor_uri,
    get_workflows_io,
)
from cmem.cmempy.workspace.activities import (
    ACTIVITY_TYPE_EXECUTE_DEFAULTWORKFLOW,
    VALID_ACTIVITY_STATUS,
)
from cmem.cmempy.workspace.activities.taskactivities import get_activities_status
from cmem.cmempy.workspace.activities.taskactivity import get_activity_status, start_task_activity
from cmem.cmempy.workspace.projects.project import get_projects
from cmem.cmempy.workspace.search import list_items
from requests import Response
from requests.status_codes import codes

from cmem_cmemc import completion
from cmem_cmemc.commands import CmemcCommand, CmemcGroup
from cmem_cmemc.commands.scheduler import scheduler
from cmem_cmemc.context import ApplicationContext

WORKFLOW_FILTER_TYPES = sorted(["project", "regex", "tag", "io"])
WORKFLOW_LIST_FILTER_HELP_TEXT = (
    "List workflows based on metadata. First parameter --filter"
    f" CHOICE can be one of {WORKFLOW_FILTER_TYPES!s}."
    " The second parameter is based on CHOICE."
)

IO_WARNING_NO_RESULT = "The workflow was executed but produced no result."
IO_WARNING_NO_OUTPUT_DEFINED = "The workflow was executed, a result was " "received but dropped."

MIME_CSV = "application/x-plugin-csv"
MIME_XLS = "application/x-plugin-excel"
MIME_NT = "application/n-triples"
MIME_JSON = "application/x-plugin-json"
MIME_XML = "application/xml"
MIME_FILE = "application/octet-stream"
MIME_ZIP = "application/x-plugin-multiCsv"
MIME_ALIGNMENT = "text/alignment"
MIME_TEXT = "text/plain"

VALID_OUTPUT_EXTENSIONS = {
    ".csv": MIME_CSV,
    ".xlsx": MIME_XLS,
    ".nt": MIME_NT,
    ".ttl": MIME_NT,
    ".json": MIME_JSON,
    ".xml": MIME_XML,
}

VALID_INPUT_EXTENSIONS = {
    ".csv": MIME_CSV,
    ".json": MIME_JSON,
    ".xml": MIME_XML,
    ".xlsx": MIME_XLS,
    ".file": MIME_FILE,
    ".zip": MIME_ZIP,
    ".txt": MIME_TEXT,
}


def is_supported_mime_type(mime_type: str) -> bool:
    """Boolean to determine if a request is multipart or not"""
    supported_mime_types = get_resource_based_dataset_types()
    return any(supported_mime_type in mime_type for supported_mime_type in supported_mime_types)


def _get_workflow_tag_labels(workflow_: dict) -> list:
    """Output a list of tag labels from a single workflow."""
    return [_["label"] for _ in workflow_["tags"]]


def _get_workflows_filtered_by_io_feature(workflows: list[dict], feature: str) -> list[dict]:
    """Get workflows filtered by io feature.

    Args:
    ----
        workflows (list): list of workflows
        feature (str): one of input-only|output-only|input-output or any

    Returns:
    -------
        list of filtered workflows form the list_items API call

    Raises:
    ------
        ValueError

    """
    possible_io_filter_values = ("input-only", "output-only", "input-output", "any")
    if feature not in possible_io_filter_values:
        raise ValueError(
            f"{feature} is an unknown filter value. " f"Use one of {possible_io_filter_values}."
        )
    filtered_workflows_ids = []
    for _ in get_workflows_io():
        ins = len(_["variableInputs"])
        outs = len(_["variableOutputs"])
        if feature == "any" and (ins == 1 or outs == 1):
            filtered_workflows_ids.append(_["projectId"] + ":" + _["id"])
            continue
        if feature == "input-only" and (ins == 1 and outs == 0):
            filtered_workflows_ids.append(_["projectId"] + ":" + _["id"])
            continue
        if feature == "output-only" and (ins == 0 and outs == 1):
            filtered_workflows_ids.append(_["projectId"] + ":" + _["id"])
            continue
        if feature == "input-output" and (ins == 1 and outs == 1):
            filtered_workflows_ids.append(_["projectId"] + ":" + _["id"])
            continue
    return [_ for _ in workflows if _["projectId"] + ":" + _["id"] in filtered_workflows_ids]


def _get_workflows_filtered(workflows: list, filter_name: str, filter_value: str) -> list:
    """Get workflows filtered according to filter name and value.

    Args:
    ----
        workflows (list): list of workflows
        filter_name (str): one of "project" or "io"
        filter_value (str): value according to fileter

    Returns:
    -------
        list of filtered workflows from the list_items API call

    Raises:
    ------
        ValueError

    """
    if filter_name not in WORKFLOW_FILTER_TYPES:
        raise ValueError(
            f"{filter_name} is an unknown filter name. " f"Use one of {WORKFLOW_FILTER_TYPES}."
        )
    # filter by regex on the label
    if filter_name == "regex":
        return [_ for _ in workflows if re.search(filter_value, _["label"])]
    # filter by project ID
    if filter_name == "project":
        return [_ for _ in workflows if _["projectId"] == filter_value]
    # filter by tag label
    if filter_name == "tag":
        return [_ for _ in workflows if filter_value in _get_workflow_tag_labels(_)]
    # filter by io feature
    if filter_name == "io":
        return _get_workflows_filtered_by_io_feature(workflows, filter_value)
    # default is unfiltered
    return workflows


def _io_check_request(info: dict, input_file: str, output_file: str, output_mimetype: str) -> None:
    """Check the requested io execution

    Raise ValueError in multiple cases.
    """
    if len(info["variableInputs"]) == 0 and input_file:
        raise ValueError(
            "You are trying to send data to a workflow without a variable "
            "input. Please remove the '-i' parameter."
        )
    if len(info["variableOutputs"]) == 0 and output_file:
        raise ValueError(
            "You are trying to retrieve data to a workflow without a variable "
            "output. Please remove the '-o' parameter."
        )
    if len(info["variableInputs"]) == 1 and not input_file:
        raise ValueError(
            "This workflow has a defined input so you need to use the '-i' "
            "parameter to send data to it."
        )
    if len(info["variableOutputs"]) == 1 and not output_file:
        raise ValueError(
            "This workflow has a defined output so you need to use the '-o' "
            "parameter to retrieve data from it."
        )
    if output_mimetype == MIME_XLS and output_file == "-":
        raise ValueError(
            "Trying to output an excel document to stdout will fail.\n"
            "Please output to a regular file instead "
            "(workflow was not executed)."
        )


def _io_get_info(project_id: str, workflow_id: str) -> dict[str, str]:
    """Get the io info dictionary of the workflow."""
    io_workflows: list[dict[str, str]] = get_workflows_io()
    for _ in io_workflows:
        info: dict[str, str] = _
        if info["id"] == workflow_id and info["projectId"] == project_id:
            return info
    raise ValueError(
        "The given workflow does not exist or is not suitable to be executed "
        "with this command.\n"
        "An io workflow needs exactly one variable input and/or one variable "
        "output."
    )


def _io_process_response(response: Response, app: ApplicationContext, output_file: str) -> None:
    """Process the workflow response of the io command."""
    with response:
        if output_file == "-":
            for line in response.iter_lines():
                if line:
                    app.echo_info(line.decode("UTF-8"))
        else:
            with click.open_file(output_file, "wb") as file:
                for chunk in response.iter_content(chunk_size=8192):
                    file.write(chunk)


def _io_guess_output(output_file: str) -> str:
    """Guess the mime type of output file name."""
    if output_file == "-":
        raise ValueError("Output mime-type not guessable, please use the --output-mimetype option.")
    file_extension = Path(output_file).suffix
    if file_extension in VALID_OUTPUT_EXTENSIONS and is_supported_mime_type(
        VALID_OUTPUT_EXTENSIONS[file_extension]
    ):
        return VALID_OUTPUT_EXTENSIONS[file_extension]
    valid_extensions = ", ".join(VALID_OUTPUT_EXTENSIONS.keys())
    raise ValueError(
        f"Files with the extension {file_extension} can not be generated. "
        f"Try one of {valid_extensions}"
    )


def _io_guess_input(input_file: str) -> str:
    """Guess the mime type of input file name."""
    if input_file == "-":
        raise ValueError("Input mime-type not guessable, please use the --output-mimetype option.")
    file_extension = Path(input_file).suffix
    if file_extension in VALID_INPUT_EXTENSIONS and is_supported_mime_type(
        VALID_INPUT_EXTENSIONS[file_extension]
    ):
        return VALID_INPUT_EXTENSIONS[file_extension]
    valid_extensions = ", ".join(VALID_INPUT_EXTENSIONS.keys())
    raise ValueError(
        f"Files with the extension {file_extension} can not be processed. "
        f"Try one of {valid_extensions}"
    )


def _workflows_get_ids() -> list[str]:
    """Get a list of workflow IDs."""
    ids = []
    for project_desc in get_projects():
        project_id = project_desc["name"]
        ids.extend([f"{project_id}:{workflow_id}" for workflow_id in get_workflows(project_id)])
    return ids


def _get_progress_bar_info(status: dict) -> str:
    """Get the workflow message from status response with colors."""
    if status is None:
        return ""
    if status["isRunning"]:
        return click.style(status["message"], fg="yellow")
    if status["concreteStatus"] == "Failed":
        return click.style(status["message"], fg="red")
    return click.style(status["message"], fg="green")


def _workflow_wait_until_finished(
    app: ApplicationContext,
    project_id: str,
    task_id: str,
    polling_interval: int,
    log_progress: bool,
) -> dict:
    """Poll workflow status until workflow is finished and return status."""
    progress = 0
    status: dict
    if log_progress:
        app.echo_success(message="", nl=True)
        with click.progressbar(  # type: ignore[var-annotated]
            show_eta=False,
            length=100,
            item_show_func=_get_progress_bar_info,  # type: ignore[arg-type]
        ) as progress_bar:
            while True:
                status = get_activity_status(project_id, task_id)
                if progress is not status["progress"]:
                    progress_bar.update(n_steps=status["progress"] - progress, current_item=status)
                    progress = status["progress"]
                app.echo_debug(f"{status['statusName']}({status['message']})")
                # wait until isRunning is false
                if not status["isRunning"]:
                    break
                time.sleep(polling_interval)
    else:
        while True:
            status = get_activity_status(project_id, task_id)
            app.echo_debug(f"{status['statusName']}({status['message']})")
            # wait until isRunning is false
            if not status["isRunning"]:
                break
            time.sleep(polling_interval)
    return status


def _workflow_echo_status(app: ApplicationContext, status: dict) -> None:
    """Print a colored status based on the returned JSON.

    Status can be Idle, Running, Canceling, Waiting, Finished
    isRunning is true for Running, Canceling, Waiting
    canceled only exists sometimes
    """
    # prepare human friendly relative time
    now = datetime.now(tz=timezone.utc)
    stamp = datetime.fromtimestamp(status["lastUpdateTime"] / 1000, tz=timezone.utc)
    time_ago = timeago.format(stamp, now, "en")
    status_name = status["statusName"]
    status_message = status["message"]
    # prepare message
    if status_name == status["message"]:
        message = f"{status_name} ({time_ago})"
    else:
        message = f"{status_name} ({status_message}, {time_ago})"

    if status["isRunning"]:
        if status_name in ("Running", "Canceling", "Waiting"):
            app.echo_warning(message)
            return
        raise ValueError(
            f"statusName is {status_name}, expecting one of: " "Running, Canceling or Waiting."
        )
    # not running can be Idle or Finished
    if status.get("failed"):
        app.echo_error(message, nl=True, err=False)
    elif status.get("cancelled"):
        app.echo_warning(message)
    elif status["statusName"] == "Idle":
        app.echo_info(message)
    else:
        # Finished and without failed or canceled status
        app.echo_success(message)


# pylint: disable=too-many-arguments
@click.command(cls=CmemcCommand, name="execute")
@click.option("-a", "--all", "all_", is_flag=True, help="Execute all available workflows.")
@click.option("--wait", is_flag=True, help="Wait until workflows are completed.")
@click.option(
    "--progress", is_flag=True, help="Wait until workflows are completed and show a progress bar."
)
@click.option(
    "--polling-interval",
    type=click.IntRange(min=0, max=60),
    show_default=True,
    default=1,
    help="How many seconds to wait between status polls. Status polls are"
    " cheap, so a higher polling interval is most likely not needed.",
)
@click.argument("workflow_ids", nargs=-1, type=click.STRING, shell_complete=completion.workflow_ids)
@click.pass_obj
def execute_command(  # noqa: PLR0913
    app: ApplicationContext,
    all_: bool,
    wait: bool,
    polling_interval: int,
    workflow_ids: tuple[str],
    progress: bool,
) -> None:
    """Execute workflow(s).

    With this command, you can start one or more workflows at the same time or
    in a sequence, depending on the result of the predecessor.

    Executing a workflow can be done in two ways: Without --wait just sends
    the starting signal and does not look for the workflow and its result
    (fire and forget). Starting workflows in this way, starts all given
    workflows at the same time.

    The optional --wait option starts the workflows in the same way, but also
    polls the status of a workflow until it is finished. In case of an error of
    a workflow, the next workflow is not started.
    """
    if workflow_ids == () and not all_:
        raise ValueError(
            "Either specify at least one workflow or use the"
            " --all option to execute all workflows."
        )
    workflows_to_execute = list(workflow_ids)
    all_workflow_ids = _workflows_get_ids()
    if all_:
        workflows_to_execute = all_workflow_ids

    for workflow_id in workflows_to_execute:
        if workflow_id not in all_workflow_ids:
            raise UsageError(f"Workflow '{workflow_id}' is not available.")
        project_id, task_id = workflow_id.split(":")
        app.echo_info(f"{workflow_id} ... ", nl=False)

        # before we start, we fetch the status
        status = get_activity_status(project_id, task_id)
        if not wait and not progress:
            if status["isRunning"]:
                # in case of a running workflow, we only output status
                app.echo_info("Already Running")
            else:
                # in case of simple call, we just start and forget
                start_task_activity(project_id, task_id)
                app.echo_info("Started")
        else:
            # in case of --wait or --progress, we poll the status until finished
            if status["isRunning"]:
                # in case of a running workflow, we only output status
                app.echo_info("Already Running ... ", nl=False)
            else:
                start_task_activity(project_id, task_id)
                app.echo_info("Started ... ", nl=False)

            status = _workflow_wait_until_finished(
                app, project_id, task_id, polling_interval, progress
            )
            # when we have a Finished status, we output it
            if not progress:
                _workflow_echo_status(app, status)
            # in case of failure, the following workflows are not executed
            if status["failed"]:
                sys.exit(1)


# pylint: disable=too-many-arguments
@click.command(cls=CmemcCommand, name="io")
@click.option(
    "--input",
    "-i",
    "input_file",
    type=click.Path(allow_dash=False, dir_okay=False, readable=True),
    shell_complete=completion.workflow_io_input_files,
    help="From which file the input is taken. If the workflow "
    "has no defined variable input dataset, this option is not allowed.",
)
@click.option(
    "--output",
    "-o",
    "output_file",
    type=click.Path(
        allow_dash=False,
        dir_okay=False,
        writable=True,
    ),
    shell_complete=completion.workflow_io_output_files,
    help="To which file the result is written to. Use '-' in order to output "
    "the result to stdout. If the workflow has no defined variable "
    "output dataset, this option is not allowed. Please note that the io "
    "command will not warn you on overwriting existing output files.",
)
@click.option(
    "--input-mimetype",
    help="Which input format should be processed: If not given, cmemc will "
    "try to guess the mime type based on the file extension or will "
    "fail.",
    type=click.Choice([*list(VALID_INPUT_EXTENSIONS.values()), "guess"]),
    default="guess",
)
@click.option(
    "--output-mimetype",
    help="Which output format should be requested: If not given, cmemc will "
    "try to guess the mime type based on the file extension or will "
    "fail. In case of an output to stdout, a default mime type "
    "will be used (JSON).",
    type=click.Choice([*list(VALID_OUTPUT_EXTENSIONS.values()), "guess"]),
    default="guess",
)
@click.option(
    "--autoconfig/--no-autoconfig",
    is_flag=True,
    show_default=True,
    help="Setup auto configuration of input datasets, e.g. in order "
    "to process CSV files with semicolon- instead of comma-separation.",
    default=True,
)
@click.argument("workflow_id", type=click.STRING, shell_complete=completion.workflow_io_ids)
@click.pass_obj
def io_command(  # noqa: PLR0913
    app: ApplicationContext,
    input_file: str,
    input_mimetype: str,
    output_file: str,
    output_mimetype: str,
    autoconfig: bool,
    workflow_id: str,
) -> None:
    """Execute a workflow with file input/output.

    With this command, you can execute a workflow that uses variable datasets
    as input, output or for configuration. Use the input parameter to feed
    data into the workflow. Likewise, use output for retrieval of the workflow
    result. Workflows without a variable dataset will throw an error.
    """
    project_id, task_id = workflow_id.split(":")
    if output_file and output_mimetype == "guess":
        output_mimetype = _io_guess_output(output_file)
    if input_file and input_mimetype == "guess":
        input_mimetype = _io_guess_input(input_file)

    _io_check_request(
        info=_io_get_info(project_id, task_id),
        input_file=input_file,
        output_file=output_file,
        output_mimetype=output_mimetype,
    )

    app.echo_debug(
        f"On workflow io execution:"
        f"project_name={project_id}, "
        f"task_name={task_id}, "
        f"input_file={input_file}, "
        f"input_mime_type={input_mimetype}, "
        f"output_mime_type={output_mimetype}, "
        f"auto_config={autoconfig}"
    )

    response = execute_workflow_io(
        project_name=project_id,
        task_name=task_id,
        input_file=input_file,
        input_mime_type=input_mimetype,
        output_mime_type=output_mimetype,
        auto_config=autoconfig,
    )
    app.echo_debug(
        f"Workflow response received after {response.elapsed} "
        f"with status {response.status_code}."
    )
    if response.status_code == codes.no_content:
        # empty content (204), warn if output was requested
        # this will happen only if info was wrong
        if output_file:
            app.echo_warning(IO_WARNING_NO_RESULT)
        # execution successful
        return
    if response.status_code == codes.ok and not output_file:
        # returns with content, warn if NO output was requested
        # this will happen only if info was wrong
        app.echo_warning(IO_WARNING_NO_OUTPUT_DEFINED)
        # execution successful
        return
    _io_process_response(response, app, output_file)


@click.command(cls=CmemcCommand, name="list")
@click.option(
    "--filter",
    "filter_",
    type=(str, str),
    multiple=True,
    shell_complete=completion.workflow_list_filter,
    help=WORKFLOW_LIST_FILTER_HELP_TEXT,
)
@click.option(
    "--id-only",
    is_flag=True,
    help="Lists only workflow identifier and no labels or other "
    "metadata. This is useful for piping the IDs into other commands.",
)
@click.option(
    "--raw", is_flag=True, help="Outputs raw JSON objects of workflow task search API response."
)
@click.pass_obj
def list_command(
    app: ApplicationContext, filter_: tuple[str, str], raw: bool, id_only: bool
) -> None:
    """List available workflow."""
    workflows = list_items(item_type="workflow")["results"]
    for _ in filter_:
        filter_type, filter_name = _
        workflows = _get_workflows_filtered(workflows, filter_type, filter_name)
    if raw:
        app.echo_info_json(workflows)
        return
    if id_only:
        # sort by combined project + task ID
        for _ in sorted(workflows, key=lambda k: (k["projectId"] + ":" + k["id"]).lower()):
            app.echo_info(_["projectId"] + ":" + _["id"])
        return
    # output a user table
    table = []
    for _ in workflows:
        row = [
            _["projectId"] + ":" + _["id"],
            _["label"],
        ]
        table.append(row)
    app.echo_info_table(table, headers=["Workflow ID", "Label"], sort_column=1)


@click.command(cls=CmemcCommand, name="status")
@click.option(
    "--project",
    "project_id",
    type=click.STRING,
    shell_complete=completion.project_ids,
    help="The project, from which you want to list the workflows. "
    "Project IDs can be listed with the 'project list' command.",
)
@click.option("--raw", is_flag=True, help="Output raw JSON info.")
@click.option(
    "--filter",
    "_filter",
    type=click.Choice(VALID_ACTIVITY_STATUS, case_sensitive=False),
    help="Show only workflows of a specific status.",
)
@click.argument("workflow_ids", nargs=-1, type=click.STRING, shell_complete=completion.workflow_ids)
@click.pass_obj
def status_command(
    app: ApplicationContext,
    project_id: str,
    raw: bool,
    _filter: str,
    workflow_ids: tuple[str],
) -> None:
    """Get status information of workflow(s)."""
    workflow_status = get_activities_status(
        status=_filter, project_name=project_id, activity=ACTIVITY_TYPE_EXECUTE_DEFAULTWORKFLOW
    )
    if workflow_ids:
        workflow_status = [
            _ for _ in workflow_status if f"{_['project']}:{_['task']}" in workflow_ids
        ]
    if raw:
        if len(workflow_status) == 1:
            app.echo_info_json(workflow_status[0])
        else:
            app.echo_info_json(workflow_status)
    else:
        for status in workflow_status:
            workflow_id = status["project"] + ":" + status["task"]
            app.echo_info(f"{workflow_id} ... ", nl=False)
            _workflow_echo_status(app, status)


@click.command(cls=CmemcCommand, name="open")
@click.argument("workflow_id", type=click.STRING, shell_complete=completion.workflow_ids)
@click.pass_obj
def open_command(app: ApplicationContext, workflow_id: str) -> None:
    """Open a workflow in your browser."""
    project_id, task_id = workflow_id.split(":")
    workflow_editor_uri = get_workflow_editor_uri().format(project_id, task_id)
    click.launch(workflow_editor_uri)
    app.echo_debug(workflow_editor_uri)


@click.group(cls=CmemcGroup)
def workflow() -> CmemcGroup:  # type: ignore[empty-body]
    """List, execute, status or open (io) workflows.

    Workflows are identified by a WORKFLOW_ID. The get a list of existing
    workflows, execute the list command or use tab-completion.
    The WORKFLOW_ID is a concatenation of a PROJECT_ID and a TASK_ID, such as
    `my-project:my-workflow`.
    """


workflow.add_command(execute_command)
workflow.add_command(io_command)
workflow.add_command(list_command)
workflow.add_command(status_command)
workflow.add_command(open_command)
workflow.add_command(scheduler)
